from typing import override

from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.blocks.schemas import CreateBulletedListItemBlock, CreateBulletedListItemData
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax import SyntaxRegistry


class BulletedListParser(LineParser):
    def __init__(self, syntax_registry: SyntaxRegistry, rich_text_converter: MarkdownRichTextConverter) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_bulleted_list_syntax()
        self._rich_text_converter = rich_text_converter

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return self._is_bulleted_list_line(context.line)

    def _is_bulleted_list_line(self, line: str) -> bool:
        return self._syntax.regex_pattern.match(line) is not None

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        block = await self._create_bulleted_list_block(context.line)
        if not block:
            return

        await self._process_nested_children(block, context)
        context.result_blocks.append(block)

    async def _process_nested_children(self, block: CreateBulletedListItemBlock, context: BlockParsingContext) -> None:
        child_lines = self._collect_child_lines(context)
        if not child_lines:
            return

        child_blocks = await self._parse_child_blocks(child_lines, context)
        if child_blocks:
            block.bulleted_list_item.children = child_blocks

        context.lines_consumed = len(child_lines)

    def _collect_child_lines(self, context: BlockParsingContext) -> list[str]:
        parent_indent_level = context.get_line_indentation_level()
        return context.collect_indented_child_lines(parent_indent_level)

    async def _parse_child_blocks(
        self, child_lines: list[str], context: BlockParsingContext
    ) -> list[CreateBulletedListItemBlock]:
        stripped_lines = self._remove_parent_indentation(child_lines, context)
        children_text = self._convert_lines_to_text(stripped_lines)
        return await context.parse_nested_markdown(children_text)

    def _remove_parent_indentation(self, lines: list[str], context: BlockParsingContext) -> list[str]:
        return context.strip_indentation_level(lines, levels=1)

    def _convert_lines_to_text(self, lines: list[str]) -> str:
        return "\n".join(lines)

    async def _create_bulleted_list_block(self, text: str) -> CreateBulletedListItemBlock | None:
        content = self._extract_list_content(text)
        if content is None:
            return None

        rich_text = await self._convert_to_rich_text(content)
        return self._build_block(rich_text)

    def _extract_list_content(self, text: str) -> str | None:
        match = self._syntax.regex_pattern.match(text)
        if not match:
            return None
        return match.group(2)

    async def _convert_to_rich_text(self, content: str):
        return await self._rich_text_converter.to_rich_text(content)

    def _build_block(self, rich_text) -> CreateBulletedListItemBlock:
        bulleted_list_content = CreateBulletedListItemData(rich_text=rich_text)
        return CreateBulletedListItemBlock(bulleted_list_item=bulleted_list_content)



import re
from typing import override

from notionary.blocks.rich_text.markdown_rich_text_converter import MarkdownRichTextConverter
from notionary.blocks.schemas import CreateCalloutBlock, CreateCalloutData
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax import SyntaxRegistry
from notionary.shared.models.icon import EmojiIcon


class CalloutParser(LineParser):
    DEFAULT_EMOJI = "💡"

    def __init__(self, syntax_registry: SyntaxRegistry, rich_text_converter: MarkdownRichTextConverter) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_callout_syntax()
        self._pattern = self._syntax.regex_pattern
        self._rich_text_converter = rich_text_converter

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        return self._pattern.search(context.line) is not None

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        block = await self._create_callout_block(context.line)
        if not block:
            return

        await self._process_nested_children(block, context)

        if self._is_nested_in_parent_context(context):
            context.parent_stack[-1].add_child_block(block)
        else:
            context.result_blocks.append(block)

    async def _process_nested_children(self, block: CreateCalloutBlock, context: BlockParsingContext) -> None:
        child_lines = self._collect_child_lines(context)
        if not child_lines:
            return

        child_blocks = await self._parse_child_blocks(child_lines, context)
        if child_blocks:
            block.callout.children = child_blocks

        context.lines_consumed = len(child_lines)

    def _collect_child_lines(self, context: BlockParsingContext) -> list[str]:
        parent_indent_level = context.get_line_indentation_level()
        return context.collect_indented_child_lines(parent_indent_level)

    async def _parse_child_blocks(
        self, child_lines: list[str], context: BlockParsingContext
    ) -> list[CreateCalloutBlock]:
        stripped_lines = self._remove_parent_indentation(child_lines, context)
        children_text = self._convert_lines_to_text(stripped_lines)
        return await context.parse_nested_markdown(children_text)

    def _remove_parent_indentation(self, lines: list[str], context: BlockParsingContext) -> list[str]:
        return context.strip_indentation_level(lines, levels=1)

    def _convert_lines_to_text(self, lines: list[str]) -> str:
        return "\n".join(lines)

    async def _create_callout_block(self, line: str) -> CreateCalloutBlock | None:
        match = self._pattern.search(line)
        if not match:
            return None

        content, emoji = self._extract_content_and_emoji(match)
        rich_text = await self._convert_to_rich_text(content)
        return self._build_block(rich_text, emoji)

    def _extract_content_and_emoji(self, match: re.Match[str]) -> tuple[str, str]:
        inline_content = match.group(1)
        if inline_content:
            return inline_content.strip(), match.group(2) or self.DEFAULT_EMOJI

        block_content = match.group(3)
        if block_content:
            return block_content.strip(), match.group(4) or self.DEFAULT_EMOJI

        return "", self.DEFAULT_EMOJI

    async def _convert_to_rich_text(self, content: str):
        return await self._rich_text_converter.to_rich_text(content)

    def _build_block(self, rich_text, emoji: str) -> CreateCalloutBlock:
        callout_data = CreateCalloutData(
            rich_text=rich_text,
            icon=EmojiIcon(emoji=emoji),
            children=[],
        )
        return CreateCalloutBlock(callout=callout_data)

    def _is_nested_in_parent_context(self, context: BlockParsingContext) -> bool:
        return bool(context.parent_stack)



import re
from dataclasses import dataclass
from enum import StrEnum


class SyntaxRegistryKey(StrEnum):
    AUDIO = "audio"
    BOOKMARK = "bookmark"
    IMAGE = "image"
    VIDEO = "video"
    FILE = "file"
    PDF = "pdf"

    # List blocks
    BULLETED_LIST = "bulleted_list"
    NUMBERED_LIST = "numbered_list"
    TO_DO = "todo"
    TO_DO_DONE = "todo_done"

    TOGGLE = "toggle"
    TOGGLEABLE_HEADING = "toggleable_heading"
    CALLOUT = "callout"
    QUOTE = "quote"
    CODE = "code"

    COLUMN_LIST = "column_list"
    COLUMN = "column"

    # Heading blocks
    HEADING_1 = "heading_1"
    HEADING_2 = "heading_2"
    HEADING_3 = "heading_3"
    HEADING = "heading"  # Shared pattern for regular headings

    DIVIDER = "divider"
    BREADCRUMB = "breadcrumb"
    TABLE_OF_CONTENTS = "table_of_contents"
    EQUATION = "equation"
    EMBED = "embed"
    TABLE = "table"
    TABLE_ROW = "table_row"

    CAPTION = "caption"
    SPACE = "space"


# some elemente need closing delimiters, others not
# either use union type or validate config in service
@dataclass(frozen=True)
class SyntaxDefinition:
    """
    Defines the syntax pattern for a block type.

    Attributes:
        start_delimiter: The opening delimiter (e.g., "```", "+++", ">")
        end_delimiter: The optional closing delimiter (empty string if none)
        regex_pattern: The compiled regex pattern to match this syntax
        end_regex_pattern: Optional compiled regex pattern for end delimiter
        is_multiline_block: Whether this block can contain child blocks
        is_inline: Whether this is an inline syntax (like [audio](url))
    """

    start_delimiter: str
    end_delimiter: str
    regex_pattern: re.Pattern
    end_regex_pattern: re.Pattern | None = None



from enum import StrEnum
from typing import Literal, Self

from pydantic import BaseModel

from notionary.shared.models.file import ExternalFile


class CoverType(StrEnum):
    EXTERNAL = "external"
    FILE = "file"


class NotionCover(BaseModel):
    type: Literal[CoverType.EXTERNAL, CoverType.FILE] = CoverType.EXTERNAL
    external: ExternalFile | None = None

    @classmethod
    def from_url(cls, url: str) -> Self:
        return cls(external=ExternalFile(url=url))



from .data_source.service import NotionDataSource
from .database.service import NotionDatabase
from .page.content.markdown.builder import MarkdownBuilder
from .page.service import NotionPage
from .workspace import NotionWorkspace, NotionWorkspaceQueryConfigBuilder

__all__ = [
    "MarkdownBuilder",
    "NotionDataSource",
    "NotionDatabase",
    "NotionPage",
    "NotionWorkspace",
    "NotionWorkspaceQueryConfigBuilder",
]



from typing import override

from notionary.blocks.schemas import Block, BlockType
from notionary.page.content.renderer.renderers.captioned_block import CaptionedBlockRenderer


class FileRenderer(CaptionedBlockRenderer):
    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.FILE

    @override
    async def _render_main_content(self, block: Block) -> str:
        url = self._extract_file_url(block)

        if not url:
            return ""

        syntax = self._syntax_registry.get_file_syntax()
        return f"{syntax.start_delimiter}{url}{syntax.end_delimiter}"

    def _extract_file_url(self, block: Block) -> str:
        if not block.file:
            return ""

        if block.file.external:
            return block.file.external.url or ""
        elif block.file.file:
            return block.file.file.url or ""

        return ""

    def _extract_file_name(self, block: Block) -> str:
        if not block.file:
            return ""

        if block.file.name:
            return block.file.name or ""

        return ""



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.syntax import SyntaxRegistry


class TableOfContentsMarkdownNode(MarkdownNode):
    def __init__(self, syntax_registry: SyntaxRegistry | None = None) -> None:
        super().__init__(syntax_registry=syntax_registry)

    @override
    def to_markdown(self) -> str:
        toc_syntax = self._syntax_registry.get_table_of_contents_syntax()
        return toc_syntax.start_delimiter



from typing import cast

from notionary.blocks.rich_text.rich_text_markdown_converter import RichTextToMarkdownConverter
from notionary.shared.entity.schemas import Describable, EntityResponseDto, Titled
from notionary.shared.models.cover import CoverType
from notionary.shared.models.icon import IconType
from notionary.shared.models.parent import DatabaseParent, DataSourceParent, ParentType


def extract_emoji_icon_from_dto(entity_dto: EntityResponseDto) -> str | None:
    if not entity_dto.icon or entity_dto.icon.type != IconType.EMOJI:
        return None
    return entity_dto.icon.emoji


def extract_external_icon_url_from_dto(entity_dto: EntityResponseDto) -> str | None:
    if not entity_dto.icon or entity_dto.icon.type != IconType.EXTERNAL:
        return None
    return entity_dto.icon.external.url if entity_dto.icon.external else None


def extract_cover_image_url_from_dto(entity_dto: EntityResponseDto) -> str | None:
    if not entity_dto.cover or entity_dto.cover.type != CoverType.EXTERNAL:
        return None
    return entity_dto.cover.external.url if entity_dto.cover.external else None


def extract_database_id(entity_dto: EntityResponseDto) -> str | None:
    if entity_dto.parent.type == ParentType.DATA_SOURCE_ID:
        data_source_parent = cast(DataSourceParent, entity_dto.parent)
        return data_source_parent.database_id if data_source_parent else None

    if entity_dto.parent.type == ParentType.DATABASE_ID:
        database_parent = cast(DatabaseParent, entity_dto.parent)
        return database_parent.database_id if database_parent else None

    return None


async def extract_title(
    entity: Titled,
    rich_text_converter: RichTextToMarkdownConverter,
) -> str:
    return await rich_text_converter.to_markdown(entity.title)


async def extract_description(
    entity: Describable,
    rich_text_converter: RichTextToMarkdownConverter,
) -> str | None:
    if not entity.description:
        return None
    return await rich_text_converter.to_markdown(entity.description)



from enum import StrEnum
from typing import Literal, Self

from pydantic import BaseModel

from notionary.shared.models.file import ExternalFile


class IconType(StrEnum):
    EMOJI = "emoji"
    EXTERNAL = "external"


class EmojiIcon(BaseModel):
    type: Literal[IconType.EMOJI] = IconType.EMOJI
    emoji: str


class ExternalIcon(BaseModel):
    type: Literal[IconType.EXTERNAL] = IconType.EXTERNAL
    external: ExternalFile

    @classmethod
    def from_url(cls, url: str) -> Self:
        return cls(external=ExternalFile(url=url))


Icon = EmojiIcon | ExternalIcon



from __future__ import annotations

from contextvars import ContextVar
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from notionary.file_upload import FileUploadHttpClient


@dataclass(frozen=True)
class PageContextProvider:
    """Context object that provides dependencies for block conversion operations."""

    page_id: str
    file_upload_client: FileUploadHttpClient


# Context variable
_page_context: ContextVar[PageContextProvider | None] = ContextVar("page_context", default=None)


def get_page_context() -> PageContextProvider:
    """Get current page context or raise if not available."""
    context = _page_context.get()
    if context is None:
        raise RuntimeError("No page context available. Use 'async with page_context(...)'")
    return context


class page_context:
    def __init__(self, provider: PageContextProvider) -> None:
        self.provider = provider
        self._token = None

    def _set_context(self) -> PageContextProvider:
        self._token = _page_context.set(self.provider)
        return self.provider

    def _reset_context(self) -> None:
        """Helper to reset context."""
        if self._token is not None:
            _page_context.reset(self._token)

    async def __aenter__(self) -> PageContextProvider:
        return self._set_context()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._reset_context()
        return False



from enum import StrEnum
from typing import Literal, Self

from pydantic import BaseModel


class FileType(StrEnum):
    EXTERNAL = "external"


class ExternalFile(BaseModel):
    url: str


class ExternalRessource(BaseModel):
    type: Literal[FileType.EXTERNAL] = FileType.EXTERNAL
    external: ExternalFile

    @classmethod
    def from_url(cls, url: str) -> Self:
        return cls(external=ExternalFile(url=url))



import re

from notionary.page.content.syntax.grammar import MarkdownGrammar
from notionary.page.content.syntax.models import SyntaxDefinition, SyntaxRegistryKey


class SyntaxRegistry:
    def __init__(self, markdown_markdown_grammar: MarkdownGrammar | None = None) -> None:
        self._markdown_grammar = markdown_markdown_grammar or MarkdownGrammar()

        self._definitions: dict[SyntaxRegistryKey, SyntaxDefinition] = {}
        self._register_defaults()

    def get_audio_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.AUDIO]

    def get_bookmark_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.BOOKMARK]

    def get_breadcrumb_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.BREADCRUMB]

    def get_bulleted_list_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.BULLETED_LIST]

    def get_callout_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.CALLOUT]

    def get_code_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.CODE]

    def get_column_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.COLUMN]

    def get_column_list_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.COLUMN_LIST]

    def get_divider_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.DIVIDER]

    def get_embed_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.EMBED]

    def get_equation_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.EQUATION]

    def get_file_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.FILE]

    def get_image_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.IMAGE]

    def get_numbered_list_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.NUMBERED_LIST]

    def get_pdf_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.PDF]

    def get_quote_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.QUOTE]

    def get_table_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.TABLE]

    def get_table_row_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.TABLE_ROW]

    def get_table_of_contents_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.TABLE_OF_CONTENTS]

    def get_todo_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.TO_DO]

    def get_todo_done_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.TO_DO_DONE]

    def get_toggle_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.TOGGLE]

    def get_toggleable_heading_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.TOGGLEABLE_HEADING]

    def get_video_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.VIDEO]

    def get_caption_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.CAPTION]

    def get_space_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.SPACE]

    def get_heading_syntax(self) -> SyntaxDefinition:
        return self._definitions[SyntaxRegistryKey.HEADING]

    def _register_defaults(self) -> None:
        self._register_audio_syntax()
        self._register_bookmark_syntax()
        self._register_image_syntax()
        self._register_video_syntax()
        self._register_file_syntax()
        self._register_pdf_syntax()

        # List blocks
        self._register_bulleted_list_syntax()
        self._register_numbered_list_syntax()
        self._register_todo_syntax()
        self._register_todo_done_syntax()

        # Container blocks
        self._register_toggle_syntax()
        self._register_toggleable_heading_syntax()
        self._register_callout_syntax()
        self._register_quote_syntax()
        self._register_code_syntax()

        # Column layout blocks
        self._register_column_list_syntax()
        self._register_column_syntax()

        self._register_heading_1_syntax()
        self._register_heading_2_syntax()
        self._register_heading_3_syntax()
        self._register_heading_syntax()  # Shared pattern for regular headings

        self._register_divider_syntax()
        self._register_breadcrumb_syntax()
        self._register_table_of_contents_syntax()
        self._register_equation_syntax()
        self._register_embed_syntax()
        self._register_table_syntax()
        self._register_table_row_syntax()

        # Post-processing and utility blocks
        self._register_caption_syntax()
        self._register_space_syntax()

    def _register_audio_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[audio](",
            end_delimiter=")",
            regex_pattern=re.compile(r"\[audio\]\(([^)]+)\)"),
        )
        self._definitions[SyntaxRegistryKey.AUDIO] = definition

    def _register_bookmark_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[bookmark](",
            end_delimiter=")",
            regex_pattern=re.compile(r"\[bookmark\]\((https?://[^\s\"]+)\)"),
        )
        self._definitions[SyntaxRegistryKey.BOOKMARK] = definition

    def _register_breadcrumb_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[breadcrumb]",
            end_delimiter="",
            regex_pattern=re.compile(r"^\[breadcrumb\]\s*$", re.IGNORECASE),
        )
        self._definitions[SyntaxRegistryKey.BREADCRUMB] = definition

    def _register_bulleted_list_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="- ",
            end_delimiter="",
            regex_pattern=re.compile(r"^(\s*)-\s+(?!\[[ xX]\])(.+)$"),
        )
        self._definitions[SyntaxRegistryKey.BULLETED_LIST] = definition

    def _register_callout_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[callout]",
            end_delimiter=")",
            regex_pattern=re.compile(
                r'\[callout\](?:\(([^")]+?)(?:\s+"([^"]+)")?\)|(?:\s+([^"\n]+?)(?:\s+"([^"]+)")?)(?:\n|$))'
            ),
        )
        self._definitions[SyntaxRegistryKey.CALLOUT] = definition

    def _register_code_syntax(self) -> None:
        code_delimiter = "```"
        definition = SyntaxDefinition(
            start_delimiter=code_delimiter,
            end_delimiter=code_delimiter,
            regex_pattern=re.compile("^" + re.escape(code_delimiter) + r"(\w*)\s*$"),
            end_regex_pattern=re.compile("^" + re.escape(code_delimiter) + r"\s*$"),
        )
        self._definitions[SyntaxRegistryKey.CODE] = definition

    def _register_column_syntax(self) -> None:
        delimiter = self._markdown_grammar.column_delimiter
        definition = SyntaxDefinition(
            start_delimiter=f"{delimiter} column",
            end_delimiter=delimiter,
            regex_pattern=re.compile(
                rf"^{re.escape(delimiter)}\s*column(?:\s+(0?\.\d+|1(?:\.0?)?))??\s*$",
                re.IGNORECASE | re.MULTILINE,
            ),
            end_regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s*$", re.MULTILINE),
        )
        self._definitions[SyntaxRegistryKey.COLUMN] = definition

    def _register_column_list_syntax(self) -> None:
        delimiter = self._markdown_grammar.column_delimiter
        definition = SyntaxDefinition(
            start_delimiter=f"{delimiter} columns",
            end_delimiter=delimiter,
            regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s*columns?\s*$", re.IGNORECASE),
            end_regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s*$"),
        )
        self._definitions[SyntaxRegistryKey.COLUMN_LIST] = definition

    def _register_divider_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="---",
            end_delimiter="",
            regex_pattern=re.compile(r"^\s*-{3,}\s*$"),
        )
        self._definitions[SyntaxRegistryKey.DIVIDER] = definition

    def _register_embed_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[embed](",
            end_delimiter=")",
            regex_pattern=re.compile(r"\[embed\]\((https?://[^\s)]+)\)"),
        )
        self._definitions[SyntaxRegistryKey.EMBED] = definition

    def _register_equation_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="$$",
            end_delimiter="$$",
            regex_pattern=re.compile(r"^\$\$\s*$"),
        )
        self._definitions[SyntaxRegistryKey.EQUATION] = definition

    def _register_file_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[file](",
            end_delimiter=")",
            regex_pattern=re.compile(r"\[file\]\(([^)]+)\)"),
        )
        self._definitions[SyntaxRegistryKey.FILE] = definition

    def _register_heading_1_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="# ",
            end_delimiter="",
            regex_pattern=re.compile(r"^#\s+(.+)$"),
        )
        self._definitions[SyntaxRegistryKey.HEADING_1] = definition

    def _register_heading_2_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="## ",
            end_delimiter="",
            regex_pattern=re.compile(r"^#{2}\s+(.+)$"),
        )
        self._definitions[SyntaxRegistryKey.HEADING_2] = definition

    def _register_heading_3_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="### ",
            end_delimiter="",
            regex_pattern=re.compile(r"^#{3}\s+(.+)$"),
        )
        self._definitions[SyntaxRegistryKey.HEADING_3] = definition

    def _register_image_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[image](",
            end_delimiter=")",
            regex_pattern=re.compile(r"(?<!!)\[image\]\(([^)]+)\)"),
        )
        self._definitions[SyntaxRegistryKey.IMAGE] = definition

    def _register_numbered_list_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="1. ",
            end_delimiter="",
            regex_pattern=re.compile(r"^(\s*)(\d+)\.\s+(.+)$"),
        )
        self._definitions[SyntaxRegistryKey.NUMBERED_LIST] = definition

    def _register_pdf_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[pdf](",
            end_delimiter=")",
            regex_pattern=re.compile(r"\[pdf\]\(([^)]+)\)"),
        )
        self._definitions[SyntaxRegistryKey.PDF] = definition

    def _register_quote_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="> ",
            end_delimiter="",
            regex_pattern=re.compile(r"^>(?!>)\s*(.+)$"),
        )
        self._definitions[SyntaxRegistryKey.QUOTE] = definition

    def _register_table_syntax(self) -> None:
        delimiter = self._markdown_grammar.table_delimiter
        definition = SyntaxDefinition(
            start_delimiter=delimiter,
            end_delimiter="",
            regex_pattern=re.compile(rf"^\s*{re.escape(delimiter)}(.+){re.escape(delimiter)}\s*$"),
        )
        self._definitions[SyntaxRegistryKey.TABLE] = definition

    def _register_table_row_syntax(self) -> None:
        delimiter = self._markdown_grammar.table_delimiter
        definition = SyntaxDefinition(
            start_delimiter=delimiter,
            end_delimiter="",
            regex_pattern=re.compile(rf"^\s*{re.escape(delimiter)}([\s\-:|]+){re.escape(delimiter)}\s*$"),
        )
        self._definitions[SyntaxRegistryKey.TABLE_ROW] = definition

    def _register_table_of_contents_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[toc]",
            end_delimiter="",
            regex_pattern=re.compile(r"^\[toc\]$", re.IGNORECASE),
        )
        self._definitions[SyntaxRegistryKey.TABLE_OF_CONTENTS] = definition

    def _register_todo_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="- [ ]",
            end_delimiter="",
            regex_pattern=re.compile(r"^\s*-\s+\[ \]\s+(.+)$"),
        )
        self._definitions[SyntaxRegistryKey.TO_DO] = definition

    def _register_todo_done_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="- [x]",
            end_delimiter="",
            regex_pattern=re.compile(r"^\s*-\s+\[x\]\s+(.+)$", re.IGNORECASE),
        )
        self._definitions[SyntaxRegistryKey.TO_DO_DONE] = definition

    def _register_toggle_syntax(self) -> None:
        delimiter = self._markdown_grammar.toggle_delimiter
        definition = SyntaxDefinition(
            start_delimiter=delimiter,
            end_delimiter=delimiter,
            regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s+(.+)$"),
            end_regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s*$"),
        )
        self._definitions[SyntaxRegistryKey.TOGGLE] = definition

    def _register_toggleable_heading_syntax(self) -> None:
        delimiter = self._markdown_grammar.toggle_delimiter
        escaped_delimiter = re.escape(delimiter)
        definition = SyntaxDefinition(
            start_delimiter=f"{delimiter} #",
            end_delimiter=delimiter,
            regex_pattern=re.compile(rf"^{escaped_delimiter}\s*(?P<level>#{{1,3}})(?!#)\s*(.+)$", re.IGNORECASE),
            end_regex_pattern=re.compile(rf"^{escaped_delimiter}\s*$"),
        )
        self._definitions[SyntaxRegistryKey.TOGGLEABLE_HEADING] = definition

    def _register_video_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[video](",
            end_delimiter=")",
            regex_pattern=re.compile(r"\[video\]\(([^)]+)\)"),
        )
        self._definitions[SyntaxRegistryKey.VIDEO] = definition

    def _register_caption_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[caption]",
            end_delimiter="",
            regex_pattern=re.compile(r"^\[caption\]\s+(\S.*)$"),
        )
        self._definitions[SyntaxRegistryKey.CAPTION] = definition

    def _register_space_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="[space]",
            end_delimiter="",
            regex_pattern=re.compile(r"^\[space\]\s*$"),
        )
        self._definitions[SyntaxRegistryKey.SPACE] = definition

    def _register_heading_syntax(self) -> None:
        definition = SyntaxDefinition(
            start_delimiter="#",
            end_delimiter="",
            regex_pattern=re.compile(r"^(#{1,3})[ \t]+(.+)$"),
        )
        self._definitions[SyntaxRegistryKey.HEADING] = definition



from typing import override

from notionary.blocks.rich_text.rich_text_markdown_converter import RichTextToMarkdownConverter
from notionary.blocks.schemas import Block, BlockType
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer
from notionary.page.content.syntax import SyntaxRegistry


class BulletedListRenderer(BlockRenderer):
    def __init__(
        self,
        syntax_registry: SyntaxRegistry | None = None,
        rich_text_markdown_converter: RichTextToMarkdownConverter | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self._rich_text_markdown_converter = rich_text_markdown_converter or RichTextToMarkdownConverter()

    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.BULLETED_LIST_ITEM

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        markdown = await self._convert_bulleted_list_to_markdown(context.block)

        if not markdown:
            context.markdown_result = ""
            return

        syntax = self._syntax_registry.get_bulleted_list_syntax()
        list_item_markdown = f"{syntax.start_delimiter}{markdown}"

        if context.indent_level > 0:
            list_item_markdown = context.indent_text(list_item_markdown)

        children_markdown = await context.render_children_with_additional_indent(1)

        if children_markdown:
            context.markdown_result = f"{list_item_markdown}\n{children_markdown}"
        else:
            context.markdown_result = list_item_markdown

    async def _convert_bulleted_list_to_markdown(self, block: Block) -> str | None:
        if not block.bulleted_list_item or not block.bulleted_list_item.rich_text:
            return None

        return await self._rich_text_markdown_converter.to_markdown(block.bulleted_list_item.rich_text)



from __future__ import annotations

from collections.abc import AsyncIterator, Callable
from typing import TYPE_CHECKING, Self

from notionary.user.service import UserService
from notionary.workspace.query.builder import NotionWorkspaceQueryConfigBuilder
from notionary.workspace.query.models import WorkspaceQueryConfig, WorkspaceQueryObjectType
from notionary.workspace.query.service import WorkspaceQueryService

if TYPE_CHECKING:
    from notionary.data_source.service import NotionDataSource
    from notionary.page.service import NotionPage
    from notionary.user import BotUser, PersonUser

type _QueryConfigInput = (
    WorkspaceQueryConfig | Callable[[NotionWorkspaceQueryConfigBuilder], NotionWorkspaceQueryConfigBuilder]
)


class NotionWorkspace:
    def __init__(
        self,
        name: str | None = None,
        query_service: WorkspaceQueryService | None = None,
        user_service: UserService | None = None,
    ) -> None:
        self._name = name
        self._query_service = query_service or WorkspaceQueryService()
        self._user_service = user_service or UserService()

    @classmethod
    async def from_current_integration(cls) -> Self:
        from notionary.user import BotUser

        bot_user = await BotUser.from_current_integration()

        return cls(name=bot_user.workspace_name)

    @property
    def name(self) -> str:
        return self._name

    async def get_pages(
        self,
        config: _QueryConfigInput | None = None,
    ) -> list[NotionPage]:
        query_config = self._resolve_config(config, WorkspaceQueryObjectType.PAGE)
        return await self._query_service.get_pages(query_config)

    async def get_pages_stream(
        self,
        config: _QueryConfigInput | None = None,
    ) -> AsyncIterator[NotionPage]:
        query_config = self._resolve_config(config, WorkspaceQueryObjectType.PAGE)
        async for page in self._query_service.get_pages_stream(query_config):
            yield page

    async def get_data_sources(
        self,
        config: _QueryConfigInput | None = None,
    ) -> list[NotionDataSource]:
        query_config = self._resolve_config(config, WorkspaceQueryObjectType.DATA_SOURCE)
        return await self._query_service.get_data_sources(query_config)

    async def get_data_sources_stream(
        self,
        config: _QueryConfigInput | None = None,
    ) -> AsyncIterator[NotionDataSource]:
        query_config = self._resolve_config(config, WorkspaceQueryObjectType.DATA_SOURCE)
        async for data_source in self._query_service.get_data_sources_stream(query_config):
            yield data_source

    def _resolve_config(
        self,
        config: _QueryConfigInput | None,
        expected_object_type: WorkspaceQueryObjectType,
    ) -> WorkspaceQueryConfig:
        if config is None:
            return self._create_default_config(expected_object_type)

        if isinstance(config, WorkspaceQueryConfig):
            return self._ensure_correct_object_type(config, expected_object_type)

        return self._build_config_from_callable(config, expected_object_type)

    def _create_default_config(self, object_type: WorkspaceQueryObjectType) -> WorkspaceQueryConfig:
        return WorkspaceQueryConfig(object_type=object_type)

    def _build_config_from_callable(
        self,
        config_callable: Callable[[NotionWorkspaceQueryConfigBuilder], NotionWorkspaceQueryConfigBuilder],
        expected_object_type: WorkspaceQueryObjectType,
    ) -> WorkspaceQueryConfig:
        builder = NotionWorkspaceQueryConfigBuilder()
        config_callable(builder)
        return self._ensure_correct_object_type(builder.build(), expected_object_type)

    def _ensure_correct_object_type(
        self,
        config: WorkspaceQueryConfig,
        expected_object_type: WorkspaceQueryObjectType,
    ) -> WorkspaceQueryConfig:
        config.object_type = expected_object_type
        return config

    async def get_users(self) -> list[PersonUser]:
        return [user async for user in self._user_service.list_users_stream()]

    async def get_users_stream(self) -> AsyncIterator[PersonUser]:
        async for user in self._user_service.list_users_stream():
            yield user

    async def get_bot_users(self) -> list[BotUser]:
        return [user async for user in self._user_service.list_bot_users_stream()]

    async def get_bot_users_stream(self) -> AsyncIterator[BotUser]:
        async for user in self._user_service.list_bot_users_stream():
            yield user

    async def search_users(self, query: str) -> list[PersonUser]:
        return [user async for user in self._user_service.search_users_stream(query)]

    async def search_users_stream(self, query: str) -> AsyncIterator[PersonUser]:
        async for user in self._user_service.search_users_stream(query):
            yield user
